"""Piezo is an AMR prediction module. Logic for handling prediction values should 
    reside within the catalogues.
"""

#! /usr/bin/env python

# -*- coding: utf-8 -*-

from .catalogue import ResistanceCatalogue  # noqa: F401
